(* Ngraph, draw a graph for PC-9801    *)
(* Copyright (c) S.Ishizaka. 1990-1991 *)

(* w֐ *)

unit Math;

interface

Type Real=Double;

type Vector=array[1..8] of Real;
type Matrix=array[1..8] of Vector;

Const MPi   :Real=3.1415926535897932385;
      MExp1 :Real=2.71828182845905;
      MEuler:Real=0.57721566490153286;

function cutdown(x:real):real;
function raise2(x:real):real;
function Log(x:Real):Real;
function Pow(a,x:Real):Real;
function Pow10(x:Real):Real;
function MInv(dim:Integer;M:Matrix;var MI:Matrix):Boolean;
function MSolve(dim:Integer; a:Matrix; b:Vector; var x:Vector):Boolean;
function Gamma(x:Real;Var val:Real):Boolean;
function IcGamma(mu,x:Real;Var val:Real):Boolean;
function Erfc(x:Real;Var val:Real):Boolean;
function Beta(p,q:Real;Var val:Real):Boolean;
function Exp1(x:Real;Var val:Real):Boolean;
function JBessel(n:Integer;x:Real;Var val:Real):Boolean;
function YBessel(n:Integer;x:Real;Var val:Real):Boolean;
function Legendre(n:Integer;x:Real;var val:Real):Boolean;
function Laguer(n:Integer;alp,x:Real;var val:Real):Boolean;
function Hermite(n:Integer;x:Real;var val:Real):Boolean;
function Chebyshev(n:Integer;x:Real;var val:Real):Boolean;

implementation

function cutdown(x:real):real;
begin
         if x>=0 then cutdown:=int(x)
         else if frac(x)=0 then cutdown:=int(x)
         else cutdown:=int(x)-1;
end;

function raise2(x:real):real;
begin
         if x<0 then raise2:=int(x)
         else if frac(x)=0 then raise2:=int(x)
         else raise2:=int(x)+1;
end;

function Log(x:Real):Real;
begin
   Log:=ln(abs(x))/ln(10)
end;

function Pow10(x:Real):Real;
Begin
  Pow10:=exp(x*Ln(10));
end;

function Pow(a,x:Real):Real;
Begin
  Pow:=exp(x*Ln(Abs(a)));
end;

function MInv(dim:Integer;M:Matrix;var MI:Matrix):Boolean;
var i,j,k:Integer;
    max:Real;
    pivot:Integer;
    vec:Vector;
begin
  for i:=1 to dim do
     for j:=1 to dim do
         if i=j then MI[i][j]:=1
                else MI[i][j]:=0;
  for i:=1 to dim do begin
     max:=0;
     for j:=i to dim do
         if Abs(M[j][i])>max then begin
             pivot:=j;
             max:=Abs(M[j][i]);
         end;
     if max<=1E-300 then begin
         MInv:=False;
         Exit;
     end;
     vec:=M[pivot];
     M[pivot]:=M[i];
     M[i]:=vec;
     vec:=MI[pivot];
     MI[pivot]:=MI[i];
     MI[i]:=vec;
     for j:=dim downto 1 do MI[i][j]:=MI[i][j]/M[i][i];
     for j:=dim downto i do M[i][j]:=M[i][j]/M[i][i];
     for j:=1 to dim do if j<>i then
         for k:=dim downto 1 do MI[j][k]:=MI[j][k]-M[j][i]*MI[i][k];
     for j:=1 to dim do if j<>i then
         for k:=dim downto i do M[j][k]:=M[j][k]-M[j][i]*M[i][k];
  end;
  MInv:=True;
end;

function MSolve(dim:Integer; a:Matrix; b:Vector; var x:Vector):Boolean;
var ai:Matrix;
    i,j:Integer;
    d:Real;
begin
    if not MInv(dim,a,ai) then begin
        MSolve:=False;
        Exit;
    end;
    for i:=1 to dim do begin
        d:=0;
        for j:=1 to dim do
            d:=d+ai[i][j]*b[j];
        x[i]:=d;
    end;
    MSolve:=True;
end;

function Gamma(x:Real;Var val:Real):Boolean;
Var M,S,x1,xx:Real;
    i:Integer;
Const GammaCoe:Array[1..19] of Real=
               (-0.42278433509846714,
                -0.23309373642178674,
                 0.19109110138769151,
                -0.24552490005400017E-1,
                -0.17645244550144320E-1,
                 0.80232730222673465E-2,
                -0.80432977560424699E-3,
                -0.3608378162548E-3,
                 0.1455961421399E-3,
                -0.175458597517E-4,
                -0.25889950224E-5,
                 0.13385015466E-5,
                -0.2054743152E-6,
                -0.1595268E-9,
                 0.62756218E-8,
                -0.12736143E-8,
                 0.923397E-10,
                 0.120028E-10,
                -0.42202E-11);
begin
   Gamma:=False;
   if (x<=0) and (Frac(x)=0) or (x>57) then exit;
   M:=1;
   if 1.5<x then
      While 1.5<x do begin
          x:=x-1;
          M:=M*x;
      end
   else if x<0.5 then
      While x<0.5 do begin
          if x<=1E-300 then exit;
          M:=M/x;
          x:=x+1;
      end;
   S:=1;
   xx:=1;
   x1:=x-1;
   for i:=1 to 19 do begin
       xx:=xx*x1;
       S:=S+xx*GammaCoe[i];
   end;
   S:=S*x;
   if S=0 then exit;
   Gamma:=True;
   val:=M/S;
end;

function Exp1(x:Real;Var val:Real):Boolean;
Var x2,xx,xx2,Qexp1,Qexp2,a:Real;
    i,n:Integer;
Const xo:Array[1..18] of Real=
          (3.3,4.0,4.8,5.8,7.1,8.6,10.4,12.5,15.2,18.4,22.2,
           26.9,32.5,39.4,47.6,57.6,69.7,83.0);
    eipt:Array[1..18] of Real=
          (44.85377247567375, 35.95520078636207,
           28.55548567959244, 22.28304130069319,
           17.15083762120765, 13.52650411391756,
           10.81282973538746,  8.781902021178544,
            7.084719123528884, 5.769115302038587,
            4.728746710237675, 3.867300609441438,
            3.178040354772841, 2.606037129493851,
            2.146957943513479, 1.767357146269297,
            1.455922099180754, 1.219698244176420);
    eimt:Array[1..18] of Real=
          (24.23610327385172,  20.63456499010558,
           17.65538999222755,  14.96789725060241,
           12.50401823871386,  10.51365383249826,
            8.830924437781962,  7.443527518977505,
            6.194083826443568,  5.167187241655524,
            4.317774043309917,  3.588549395492134,
            2.987594411636582,  2.476696474779806,
            2.058451487317135,  1.706965828524471,
            1.414702600059552,  1.190641095161813);
begin
   Exp1:=False;
   x2:=Abs(x);
   if (x=0) or (x2>174) then exit;
   if x2>=90 then begin
       n:=1;
       xx:=1;
       Qexp2:=1;
       Repeat
           xx:=xx*n/x;
           Inc(n);
           Qexp1:=Qexp2;
           Qexp2:=Qexp1+xx;
       until Qexp1=Qexp2;
       Qexp1:=Qexp1/x;
       if x<0 then Qexp1:=-Qexp1*Exp(-x2)
              else Qexp1:=Qexp1*Exp(x2);
   end else if x2>=3 then begin
       i:=1;
       While (i<18) and (x2>=0.5*(xo[i]+xo[i+1])) do Inc(i);
       if (x>0) then xx:=eipt[i]/100
                else xx:=eimt[i]/100;
       xx2:=1/(-xo[i]);
       Qexp2:=xx;
       n:=1;
       Repeat
           Qexp1:=Qexp2;
           xx:=(xo[i]-x2)/n*(xx+xx2);
           if x<0 then xx:=-xx;
           Qexp2:=Qexp1+xx;
           xx2:=xx2*(x2-xo[i])/(-xo[i]);
           Inc(n);
       until (Qexp2=Qexp1);
       if x<0 then Qexp1:=-Qexp1*Exp(-x2)
              else Qexp1:=Qexp1*Exp(x2);
   end else begin
       n:=1;
       xx:=1;
       Qexp2:=0;
       Repeat
           xx:=xx*x/n;
           Qexp1:=Qexp2;
           Qexp2:=Qexp1+xx/n;
           Inc(n);
       until Qexp1=Qexp2;
       Qexp1:=Qexp1+MEuler+Ln(x2);
   end;
   Exp1:=True;
   val:=Qexp1;
end;

function IcGamma(mu,x:Real;Var val:Real):Boolean;
Var a,u,p0,p1,q0,q1:Real;
    i,i2,i3:Integer;
begin
   IcGamma:=False;
   if (x<0) or (mu<0) then exit;
   if mu=0 then begin
       if not Exp1(-x,val) then exit
                           else val:=-val;
   end else if x=0 then begin
       if not Gamma(mu,val) then exit;
   end else begin
       i2:=0;
       While mu>1 do begin
           i:=0;
           Repeat
             if i=56 then exit;
             Inc(i);
             mu:=mu-1;
           until mu<=1;
           i2:=i;
       end;
       u:=Exp(-x+mu*Ln(x));
       if x<=1 then begin
           if not Gamma(mu,p1) then exit;
           p0:=1;
           q0:=1;
           i:=0;
           Repeat
               Inc(i);
               q0:=q0*x/(mu+i);
               p0:=p0+q0;
           until (q0<=1E-17) or (i=20);
           val:=p1-u/mu*p0;
       end else begin
           i3:=Trunc(120/x+5);
           p0:=0;
           if mu<>1 then p1:=1E-78/(1-mu)
                    else p1:=1E-78;
           q0:=p1;
           q1:=x*q0;
           for i:=1 to i3 do begin
               a:=i-mu;
               p0:=p1+a*p0;
               q0:=q1+a*q0;
               p1:=x*p0+i*p1;
               q1:=x*q0+i*q1;
               if i>=50 then begin
                   q0:=q0/i;
                   p0:=p0/i;
                   q1:=q1/i;
                   p1:=p1/i;
               end;
           end;
           val:=u/q1*p1;
       end;
       for i:=1 to i2 do begin
           val:=u+mu*val;
           u:=u*x;
           mu:=mu+1;
       end;
   end;
   IcGamma:=True;
end;

function Erfc(x:Real;Var val:Real):Boolean;
begin
    if not IcGamma(0.5,x*x,val) then Erfc:=False
    else begin
        Erfc:=True;
        val:=val/Sqrt(MPi);
        if x<0 then val:=2-val;
    end;
end;

function Beta(p,q:Real;Var val:Real):Boolean;
var a,b,c:Real;
begin
    Beta:=False;
    if not Gamma(p,a) then exit;
    if not Gamma(q,b) then exit;
    if not Gamma(p+q,c) then exit;
    if Abs(c)<1E-300 then exit;
    Beta:=True;
    val:=a*b/c;
end;

function JBessel(n:Integer;x:Real;Var val:Real):Boolean;
var n2,m,l,i:Integer;
    x2,R,T1,T2,T3,S,J:Real;
begin
    JBessel:=False;
    x2:=Abs(x);
    n2:=Abs(n);
    if (n2>1000) or (x2>1000) then Exit;
    if x2<=2E-5 then begin
        T1:=x2*0.5;
        if n2=0 then J:=1-T1*T1
        else if x2<=1E-77 then J:=0
        else begin
            T2:=1;
            T3:=1;
            i:=1;
            While (i<=n2) and (Abs(T3)>Abs(T2/T1)*1E-77) do begin
                T3:=T3*T1/T2;
                T2:=T2+1;
                Inc(i);
            end;
            if (i<=n2) then J:=0
                       else J:=T3*(1-T1*T1/T2);
        end;
    end else begin
        if n2>x2 then m:=n2
                 else m:=Trunc(x2);
        if x2>=100 then l:=Trunc(0.073*x2+47)
        else if x2>=10 then l:=Trunc(0.27*x2+27)
        else if x2>1 then l:=Trunc(1.4*x2+14)
        else l:=14;
        Inc(m,l);
        T3:=0;
        T2:=1E-75;
        S:=0;
        for i:=m-1 downto 0 do begin
            T1:=2*(i+1)/x2*T2-T3;
            if i=n2 then J:=T1;
            if ((i mod 2)=0) and (i<>0) then begin
                S:=S+T1;
                if Abs(S)>=1E55 then begin
                    T1:=T1*1E-55;
                    T2:=T2*1E-55;
                    S:=S*1E-55;
                    J:=J*1E-55;
                end;
            end;
            T3:=T2;
            T2:=T1;
        end;
        S:=2*S+T1;
        J:=J/S;
    end;
    if (n<0) and ((n2 mod 2)=1) then J:=-J;
    if (x<0) and ((n2 mod 2)=1) then J:=-J;
    val:=J;
    JBessel:=True;
end;

function YBessel(n:Integer;x:Real;Var val:Real):Boolean;
var n2,m,i,l:Integer;
    x2,R,T1,T2,T3,S,SS,W,J0,J1,Y0,Y1,Y2:Real;
begin
    YBessel:=False;
    x2:=Abs(x);
    n2:=Abs(n);
    if (n2>1000) or (x2>1000) or (x2<=1E-300) then Exit;
    if x2<=2E-5 then begin
        T1:=x2*0.5;
        J0:=1-T1*T1;
        Y0:=2/MPi*((MEuler+Ln(T1))*J0+T1*T1);
        if n2<>0 then begin
            J1:=T1-0.5*T1*T1*T1;
            Y1:=(J1*Y0-2/MPi/x2)/J0;
            if n2>=-65/Log(T1) then exit;
            for i:=2 to n2 do begin
                Y2:=2*(i-1)/x2*Y1-Y0;
                Y0:=Y1;
                Y1:=Y2;
            end;
        end;
    end else begin
        if x2>=100 then m:=Trunc(1.073*x2+47)
        else if x2>=10 then m:=Trunc(1.27*x2+28)
        else if x2>=1 then m:=Trunc(2.4*x2+15)
        else if x2>=0.1 then m:=17
        else m:=10;
        T3:=0;
        T2:=1E-75;
        S:=0;
        SS:=0;
        for i:=m-1 downto 0 do begin
            T1:=2*(i+1)/x2*T2-T3;
            if ((i mod 2)=0) and (i<>0) then begin
                S:=S+T1;
                l:=i div 2;
                if (l mod 2)=0  then SS:=SS-T1/l
                                else SS:=SS+T1/l;
                if Abs(S)>=1E55 then begin
                    T1:=T1*1E-55;
                    T2:=T2*1E-55;
                    S:=S*1E-55;
                    SS:=SS*1E-55;
                end;
            end;
            T3:=T2;
            T2:=T1;
        end;
        S:=2*S+T1;
        J0:=T1/S;
        J1:=T3/S;
        SS:=SS/S;
        Y0:=2/MPi*((MEuler+Ln(x2*0.5))*J0+2*SS);
        if n2<>0 then begin
            Y1:=(J1*Y0-2/MPi/x2)/J0;
            for i:=2 to n2 do begin
                if Y1<=-1E65 then begin
                    W:=Y1*1E-10;
                    Y2:=2*(i-1)/x2*W-Y0*1E-10;
                    if Abs(Y2)>=1E65 then exit;
                    Y2:=Y2*1E10;
                end else Y2:=2*(i-1)/x2*Y1-Y0;
                Y0:=Y1;
                Y1:=Y2;
            end;
        end;
    end;
    if n=0 then val:=Y0
    else begin
        if (n<0) and ((n2 mod 2)=1) then Y1:=-Y1;
        if (x<0) and ((n2 mod 2)=1) then Y1:=-Y1;
        val:=Y1;
    end;
    YBessel:=True;
end;

function Legendre(n:Integer;x:Real;var val:Real):Boolean;
var l1,l2,tmp1,tmp2:Real;
    i:Integer;
begin
  Legendre:=False;
  if n<0 then Exit;
  l1:=1;
  if n=0 then val:=l1
  else begin
    l2:=x;
    if n=1 then val:=l2
    else begin
      for i:=2 to n do begin
        tmp1:=x*l2;
        tmp2:=tmp1-l1;
        val:=tmp2+tmp1-tmp2/i;
        l1:=l2;
        l2:=val;
      end;
    end;
  end;
  Legendre:=True;
end;

function Laguer(n:Integer;alp,x:Real;var val:Real):Boolean;
var i:Integer;
    l1,l2,tmp1,tmp2:Real;
begin
  Laguer:=False;
  if n<0 then Exit;
  l1:=1;
  if n=0 then val:=l1
  else begin
    l2:=1-x+alp;
    if n=1 then val:=l2
    else begin
      tmp1:=1+x-alp;
      tmp2:=1-alp;
      for i:=2 to n do begin
        val:=(l2-l1)+l2-(tmp1*l2-tmp2*l1)/i;
        l1:=l2;
        l2:=val;
      end;
    end;
  end;
  Laguer:=True;
end;

function Hermite(n:Integer;x:Real;var val:Real):Boolean;
var l1,l2:Real;
    i:Integer;
begin
  Hermite:=False;
  if n<0 then Exit;
  l1:=1;
  if n=0 then val:=l1
  else begin
    l2:=2*x;
    if n=1 then val:=l2
    else begin
      for i:=2 to n do begin
        val:=x*l2-(i-1)*l1;
        val:=val+val;
        l1:=l2;
        l2:=val;
      end;
    end;
  end;
  Hermite:=True;
end;

function Chebyshev(n:Integer;x:Real;var val:Real):Boolean;
var l1,l2,tmp1:Real;
    i:Integer;
begin
  Chebyshev:=False;
  if n<0 then Exit;
  l1:=1;
  if n=0 then val:=l1
  else begin
    l2:=x;
    if n=1 then val:=l2
    else begin
      tmp1:=x+x;
      for i:=2 to n do begin
        val:=tmp1*l2-l1;
        l1:=l2;
        l2:=val;
      end;
    end;
  end;
  Chebyshev:=True;
end;

end.
