(* Ngraph, draw a graph for PC-9801    *)
(* Copyright (c) S.Ishizaka. 1990-1991 *)

(* XvCԎ葱 *)

{  Procedure of Spline interpolation

  Usage | Spline(gx,gy,c1,c2,c3,datnum,bc0,bc1,df0,df1);

    bc0, bc1 : 0  [_Ŏݒ
	       1  [_̔W^        df0, df1
	       2@[_̓KW^    df0, df1
	       3  IE
}

unit PRSPLINE;

interface

uses Math;

Type Real=Double;

function Spline(var gx,gy,c1,c2,c3;datnum,bc0,bc1:Integer;df0,df1:Real):Boolean;
procedure BSpline(Edge:Integer;var x;var c);

implementation

type data=array[1..MAXINT div 8] of Real;

function Check(var d,mu,ram,gx,gy:data;datnum:Integer):Boolean;
var h1,h2:Real;
    i:Integer;
begin
   h1:=gx[2]-gx[1];
   if h1=0 then begin
       Check:=False;
       exit;
   end;
   for i:=2 to datnum-1 do begin
     h2:=gx[i+1]-gx[i];
     if (h2=0) or (h1=0) or (h1+h2=0) then begin
	 Check:=False;
	 exit;
     end;
     h1:=h2;
   end;
   Check:=True;
end;

procedure SetVal(var d,mu,ram,gx,gy:data;datnum:Integer);
var h1,h2,y1,y2:Real;
    i:Integer;
begin
   h1:=gx[2]-gx[1];
   y1:=gy[2]-gy[1];
   for i:=2 to datnum-1 do begin
     h2:=gx[i+1]-gx[i];
     y2:=gy[i+1]-gy[i];
     ram[i]:=h2/(h1+h2);
     mu[i]:=1-ram[i];
     d[i]:=6/(h1+h2)*(y2/h2-y1/h1);
     h1:=h2;
     y1:=y2;
   end;
   mu[1]:=0;
   ram[datnum]:=0;
end;

function Solve(var d,mu,ram,gx,gy:data;datnum:integer;var x:real):Boolean;
var i:Integer;
    h:Real;
begin
     x:=-mu[1]/2;
     d[1]:=d[1]/2;
     mu[1]:=ram[1]/2;
     for i:=2 to datnum do begin
       h:=2-mu[i]*mu[i-1];
       if h=0 then begin
	   Solve:=False;
	   exit;
       end;
       x:=-mu[i]*x/h;
       d[i]:=(d[i]-mu[i]*d[i-1])/h;
       mu[i]:=ram[i]/h;
     end;
     x:=x-mu[datnum];
     Solve:=True;
end;

function Solve2(var d,mu,ram,gx,gy:data;datnum:integer;var x:real):Boolean;
var i:Integer;
    h:Real;
begin
     x:=-ram[datnum]/2;
     d[datnum]:=d[datnum]/2;
     ram[datnum]:=mu[datnum]/2;
     for i:=datnum-1 downto 1 do begin
       h:=2-ram[i]*ram[i+1];
       if h=0 then begin
	   Solve2:=False;
	   exit;
       end;
       x:=-ram[i]*x/h;
       d[i]:=(d[i]-ram[i]*d[i+1])/h;
       ram[i]:=mu[i]/h;
     end;
     x:=x-ram[1];
     Solve2:=True;
end;

function Period(var d,mu,ram,gx,gy:data;datnum:integer;var df0:real):Boolean;
var h,x,a,b,c,e:Real;
    cond:Boolean;

	procedure PeriodSet;
	begin
	       SetVal(d,mu,ram,gx,gy,datnum);
	       h:=gx[2]-gx[1];
	       d[1]:=6.0/h/h*(gy[2]-gy[1]);
	       ram[1]:=1;
	       mu[1]:=6.0/h;
	       h:=gx[datnum]-gx[datnum-1];
	       d[datnum]:=-6.0/h/h*(gy[datnum]-gy[datnum-1]);
	       mu[datnum]:=1;
	       ram[datnum]:=-6.0/h;
	end;

begin
     PeriodSet;
     if not Solve(d,mu,ram,gx,gy,datnum,x) then begin
	 Period:=False;
	 exit;
     end;
     a:=d[datnum];
     b:=x;
     PeriodSet;
     if not Solve2(d,mu,ram,gx,gy,datnum,x) then begin
	 Period:=False;
	 exit;
     end;
     c:=d[1];
     e:=x;
     h:=e-b;
     if h=0 then begin
	 Period:=False;
	 exit;
     end;
     df0:=(a*e-b*c)/h;
     Period:=True;
end;

function Boundary(var d,mu,ram,gx,gy:data;datnum:integer;
		    bc0,bc1:Integer;df0,df1:Real):Boolean;
var b,coe:Vector;
    m:Matrix;
    h,y:real;
    i,j,k,st:integer;
begin
    if bc0=0 then begin
	 for i:=1 to 4 do begin
	   m[i][4]:=1;
	   b[i]:=gy[i];
	   for j:=3 downto 1 do m[i][j]:=m[i][j+1]*gx[i];
	 end;
	 if not MSolve(4,m,b,coe) then begin
	     Boundary:=False;
	     exit;
	 end;
	 df0:=3*coe[1]*m[1][2]+2*coe[2]*m[1][3]+coe[3];
    end;
    if bc1=0 then begin
	 for i:=1 to 4 do begin
	   m[i][4]:=1;
	   b[i]:=gy[datnum+i-4];
	   for j:=3 downto 1 do m[i][j]:=m[i][j+1]*gx[datnum+i-4];
	 end;
	 if not MSolve(4,m,b,coe) then begin
	     Boundary:=False;
	     exit;
	 end;
	 df1:=3*coe[1]*m[4][2]+2*coe[2]*m[4][3]+coe[3];
    end;
    if bc0=3 then begin
	if not Period(d,mu,ram,gx,gy,datnum,df0) then begin
	    Boundary:=False;
	    exit;
	end;
	bc0:=2;
	bc1:=2;
	df1:=df0;
    end;
    if bc0=2 then begin
	d[1]:=2*df0;
	ram[1]:=0;
    end else if (bc0=0) or (bc0=1) then begin
	h:=gx[2]-gx[1];
	d[1]:=6.0/h*((gy[2]-gy[1])/h-df0);
	ram[1]:=1;
    end else begin
	Boundary:=False;
	exit;
    end;
    if bc1=2 then begin
	d[datnum]:=2*df1;
	mu[datnum]:=0;
    end else if (bc1=0) or (bc1=1) then begin
	h:=gx[datnum]-gx[datnum-1];
	d[datnum]:=6.0/h*(df1-(gy[datnum]-gy[datnum-1])/h);
	mu[datnum]:=1;
    end else begin
	Boundary:=False;
	exit;
    end;
    Boundary:=True;
end;

function Spline(var gx,gy,c1,c2,c3;datnum,bc0,bc1:Integer;df0,df1:Real):Boolean;
var
    i:integer;
    h,x,m:real;
begin
   if not Check(data(c1),data(c2),data(c3),data(gx),data(gy),datnum) then begin
       Spline:=False;
       exit;
   end;
   if not Boundary(data(c1),data(c2),data(c3),data(gx),data(gy),datnum,bc0,bc1,df0,df1) then begin
       Spline:=False;
       exit;
   end;
   SetVal(data(c1),data(c2),data(c3),data(gx),data(gy),datnum);
   if not Solve(data(c1),data(c2),data(c3),data(gx),data(gy),datnum,x) then begin
       Spline:=False;
       exit;
   end;

   data(c3)[datnum]:=data(c1)[datnum];
   for i:=datnum-1 downto 1 do
     data(c3)[i]:=data(c1)[i]-data(c2)[i]*data(c3)[i+1];

   for i:=1 to datnum-1 do begin
     h:=data(gx)[i+1]-data(gx)[i];
     m:=data(c3)[i];
     data(c3)[i]:=(data(c3)[i+1]-m)/6.0/h;
     data(c2)[i]:=m/2;
     data(c1)[i]:=(data(gy)[i+1]-data(gy)[i])/h-h*(data(c2)[i]+h*data(c3)[i]);
   end;

   Spline:=True;
end;

procedure BSpline(Edge:Integer;var x;var c);
Type BS=Array[0..3] of Real;
begin
  if Edge=0 then begin
      BS(c)[0]:=(BS(x)[0]+4.0*BS(x)[1]+1.0*BS(x)[2])/6.0;
      BS(c)[1]:=(-3.0*BS(x)[0]+3.0*BS(x)[2])/6.0;
      BS(c)[2]:=(3.0*BS(x)[0]-6.0*BS(x)[1]+3.0*BS(x)[2])/6.0;
      BS(c)[3]:=(-1.0*BS(x)[0]+3.0*BS(x)[1]-3.0*BS(x)[2]+BS(x)[3])/6.0;
  end else if Edge=1 then begin
      BS(c)[0]:=(12.0*BS(x)[0])/12.0;
      BS(c)[1]:=(-36.0*BS(x)[0]+36.0*BS(x)[1])/12.0;
      BS(c)[2]:=(36.0*BS(x)[0]-54.0*BS(x)[1]+18.0*BS(x)[2])/12.0;
      BS(c)[3]:=(-12.0*BS(x)[0]+21.0*BS(x)[1]-11.0*BS(x)[2]+2.0*BS(x)[3])/12.0;
  end else if Edge=2 then begin
      BS(c)[0]:=(3.0*BS(x)[0]+7.0*BS(x)[1]+2.0*BS(x)[2])/12.0;
      BS(c)[1]:=(-9.0*BS(x)[0]+3.0*BS(x)[1]+6.0*BS(x)[2])/12.0;
      BS(c)[2]:=(9.0*BS(x)[0]-15.0*BS(x)[1]+6.0*BS(x)[2])/12.0;
      BS(c)[3]:=(-3.0*BS(x)[0]+7.0*BS(x)[1]-6.0*BS(x)[2]+2.0*BS(x)[3])/12.0;
  end else if Edge=3 then begin
      BS(c)[0]:=(2.0*BS(x)[0]+8.0*BS(x)[1]+2.0*BS(x)[2])/12.0;
      BS(c)[1]:=(-6.0*BS(x)[0]+6.0*BS(x)[2])/12.0;
      BS(c)[2]:=(6.0*BS(x)[0]-12.0*BS(x)[1]+6.0*BS(x)[2])/12.0;
      BS(c)[3]:=(-2.0*BS(x)[0]+6.0*BS(x)[1]-7.0*BS(x)[2]+3.0*BS(x)[3])/12.0;
  end else if Edge=4 then begin
      BS(c)[0]:=(2.0*BS(x)[0]+7.0*BS(x)[1]+3.0*BS(x)[2])/12.0;
      BS(c)[1]:=(-6.0*BS(x)[0]-3.0*BS(x)[1]+9.0*BS(x)[2])/12.0;
      BS(c)[2]:=(6.0*BS(x)[0]-15.0*BS(x)[1]+9.0*BS(x)[2])/12.0;
      BS(c)[3]:=(-2.0*BS(x)[0]+11.0*BS(x)[1]-21.0*BS(x)[2]+12.0*BS(x)[3])/12.0;
  end;
end;

end.
